/*
    Gn: A Server for the Internet Gopher Protocol(*).
    File: gn/gn.c
    Version 1.19
    
    Copyright (C) 1993  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    (*) Gopher is a registered trademark of the Univ. of Minn.
*/

#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "gn.h"

int		http;
char		mime_accept[MAXLEN];

static char *www_preparse();
static void get_header();

main( argc, argv)
int	argc;
char	*argv[];

{
	char	selector[PATHLEN],
		*selp = selector;
	int	itemtype;
	Item	thisitem;

	gn_init( argc, argv);

	get_input( selector);
	strcpy( thisitem.selector, selector);

	/* If the first letter is in the following list it should
	   be an HTTP request					   */

	switch ( *selector) {
		case 'G':
		case 'H':
		case 'S':
		case 'P':
		case 'T':
			if ( (selp = www_preparse( selector)) == NULL)
				exit(0);
	}
	
	parse_selector( &thisitem, selp, FULL);

	if ( chkaccess( &thisitem, LOG))
		itemtype = chkcache( &thisitem);
	else
		itemtype = DENIED;

	switch (itemtype) {
		case ROOT:
			sendcache( &thisitem);
			break;
		case TEXT:
			sendtext( &thisitem);
			break;
		case RANGE:
			sendrange( &thisitem);
			break;
		case DIR:
			sendcache( &thisitem);
			break;
		case BIN:
			sendbin( &thisitem);
			break;
		case EXEC:
			sendexec( &thisitem);
			break;
		case SEARCH:
			sendsearch( &thisitem);
			break;
		case HEADER:
			sendhead( &thisitem);
			break;
		case COMMENT:
			exit( 0);
			break;
		case DENIED:
			senderr( DENYMSG, &thisitem);
			break;
		case UNKNOWN:
			senderr( "Unknown selector type", &thisitem);
			break;
		default:
			senderr( "Unknown server error", &thisitem);
	}
	return (0);
}


static char *
www_preparse( sel)
char	*sel;
{
	register char	*cp;
	char		errmsg[BUFSIZE],
			method[SMALLLEN];
	cp = sel;
	while ( *cp && !isspace( *cp))
		cp++;
	*cp++ = '\0';
	strcpy( method, sel);
	sel = cp;
	while (  *sel && isspace( *sel))
		sel++;
	cp = sel;
	while (  *cp && !isspace( *cp))
		cp++;
	if ( *cp ) { /* There's more, check HTTP Version */
		*cp++ = '\0';
		while (  *cp && isspace( *cp))
			cp++;
	}
	http = (streq( cp, "HTTP/1.0") ? HTTP1_0 : HTTP0_9);
	if ( streq( method, "HEAD") )
		http |= HTTP1_0HEAD;
	if ( streq( method, "GET") || ( http & HTTP1_0HEAD)) {
		if ( *sel == '/' )
			sel++;

		if ( http & HTTP1_0 ) 
			get_header();

		return (sel);
	}
	else {
		strcpy( errmsg, method);
		strcat( errmsg, ": Method not implemented");
		www_err( "501", errmsg);
		sel = NULL;
	}
	return sel;
}



static void
get_header( )
{
	char	header[BUFSIZE];

	mime_accept[0] = '\0';
	while ( get_input( header) != NULL) {
		if ( strncmp( header, "Accept:", 7) == 0 )
			strcpy( mime_accept, header + 7);
		if ( *header == '\0')
			return;
	}
}
