/*
    Gn: A Server for the Internet Gopher Protocol(*).
    File: gn/gn.h
    Version 1.19
    
    Copyright (C) 1993  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    (*) Gopher is a registered trademark of the Univ. of Minn.
*/

#include <stdio.h>
#include "../config.h"

#define VERSION	"GN/1.19"
#define HTTPVERSION "HTTP/1.0"
#define	PATHLEN	(4096)
#define	MAXLEN	(4096)
#define BUFSIZE	(4096)
#define	MIDLEN	(1024)
#define SMALLLEN (256)

#define streq( a, b)	( strcmp( (a), (b)) == 0 )

#ifndef TRUE
#define TRUE	(1)
#endif

#ifndef FALSE
#define FALSE	(0)
#endif


#define DENYMSG	"Sorry, access denied or file does not exist."
#define FREE	(0)
#define ROOTCHK	(1)
#define DIRCHK	(2)
#define LOG	TRUE
#define NOLOG	FALSE
#define FULL	TRUE
#define PARTIAL	FALSE

#define ROOT	(-1)
#define	TEXT	(0)
#define	DIR	(1)
#define CSO	(2)
#define ERR	(3)
#define SEARCH	(4)
#define EXEC	(5)
#define BIN	(6)
#define RANGE	(7)
#define HEADER	(8)
#define COMMENT	(10)
#define UNKNOWN	(11)
#define	DENIED	(12)

#define NONE		(0)
#define HTTP0_9		(1)
#define HTTP1_0		(2)
#define HTTP1_0HEAD	(4)

#define	LOCAL		(1)
#define ILINE		(2)
#define REMGOPHER	(3)
#define REMGN		(4)
#define HTTPTEXT	(5)

extern char	rootdir[],
		root_title[],
		cfname[],
		afname[],
		myhost[],
		myport[],
		remotehost[],
		remaddr[],
		mime_accept[],
		*get_input(),
		*get_cache_line(),
		*strlower();

extern int	http,
		accesstype,
		chkcache(),
		chkaccess();

extern void	gn_init(),
		writelog(),
		chop(),
		sendtext(),
		sendbin(),
		senddir(),
		sendexec(),
		sendcache(),
		sendrange(),
		sendsearch(),
		sendhead(),
		cache_search(),
		parse_selector(),
		cache_index(),
		exit(),
		get_remote_info(),
		send_text_line(),
		send_cache_line(),
		www_cache_line(),
		www_unescape(),
		cache_prolog(),
		http_prolog(),
		cgi_env(),
		www_query(),
		www_err(),
		cache_epilog(),
		send_end(),
		senderr(),
		senderr2();

extern FILE	*safer_popen();

typedef struct Item {
	char	selector[PATHLEN],	/* The original selector */
		name[PATHLEN],		/* filled  by chkcache, used for www */
		gtype[PATHLEN],		/*
					 * Typically gopher type 0,1, etc.
					 * but really everything up to first
                                         * '/' in the selector
					 */
		filepath[PATHLEN],	/* Complete pathname of file  */
 		cachepath[PATHLEN],	/* Complete pathname of cache which */
					/* which contains this item */
 		accdir[PATHLEN],	/* path of access file directory */
		extra[PATHLEN],		/* Stuff after the last tab */
		args[PATHLEN],		/* Args for exec */
		content_type[MIDLEN],	/* MIME content type */
		mod_date[SMALLLEN],	/* File modification date */
		length[SMALLLEN],	/* File length */
		suffix[SMALLLEN];	/* File name suffix */

	long	range_start,		/* Start and end of text file range */
		range_end;

	int	compressed;		/* True if file is compressed */

} Item;


typedef struct Cache_entry {
	char	type1,			/*
					 * The single char type which is
					 * the first char in the .cache line
					 */
		name[MAXLEN],
		path[MAXLEN],
		host[MAXLEN],
		port[MAXLEN],
		cacheline[MAXLEN],
		content[MIDLEN],
		suffix[SMALLLEN];

	int	entrytype;		/* 
					 * This can be REMGOPHER, REMGN,
                                         * LOCAL, COMMENT, or HTTPTEXT.
					 */
} Cache_entry;

extern Cache_entry	*read_cache();
