/*
    Gn: A Server for the Internet Gopher Protocol(*).
    File: gn/sio.c
    Version 1.19
    
    Copyright (C) 1993  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    (*) Gopher is a registered trademark of the Univ. of Minn.
*/

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "gn.h"


static void	www_escape(),
		www_replace(),
		parse_cache();


char *
get_input( sel)
char	*sel;
{
	if ( fgets( sel, PATHLEN, stdin) == NULL) {
		return NULL;
	}
	chop( sel);
	return ( sel);
}


void
send_text_line( line)
char	*line;
{
	if ( strchr( line, '\n') != NULL )
		chop( line);
	if ( http) {
		fputs( line, stdout);
		fputs( "\n", stdout);
	}
	else {
		if ( streq( line, "."))
		strcat( line, ".");
			/* Double a '.' if on line by itself */
		fputs( line, stdout);
		fputs( "\r\n", stdout);
	}
}

void
send_cache_line( chp)
Cache_entry	*chp;
{
	register char	*cp,
			*cp2;
	char		line[MAXLEN];

	int	brackdepth = 0;
	Item    item;

	strcpy( line , chp->cacheline);
        if ( chp->entrytype == HTTPTEXT )
		return;

	if ( accesstype == DIRCHK) {
		if ( chp->entrytype == LOCAL ) {
			parse_selector( &item, chp->path, FULL);
			if (!chkaccess( &item, NOLOG))  
				return;
		}
	}
        if ( chp->entrytype == ILINE ) {
		/* Delete everything in <> angle brackets */
		cp = cp2 = line;
		while ( *cp2 ) {
			if ( *cp2 == '<')
				brackdepth++;
			if ( brackdepth >0 )
				cp2++;
			else 
				*cp++ = *cp2++;
			if ( *(cp2 - 1) == '>')
				brackdepth--;
		}
		*cp = '\0';
	}
	fputs( line, stdout);
	fputs( "\r\n", stdout);
}

void
send_end()
{
	fputs( ".\r\n", stdout);
}



/*
 * read_cache() reads from a cache file (fp) and fills the Cache_entry
 * struct pointed to by cep.  It puts the primary cache file line in
 * chp->cacheline first.  Then it reads the secondary line to fill
 * out the entrytype, content and suffix fields.  Next it calls
 * parse_cache( chp) which parses cep->cacheline and fills in the 
 * remaining fields.  It returns NULL when the cache file has been
 * exhausted and otherwise returns its first argument.
 */

Cache_entry *
read_cache( cep, fp)
Cache_entry	*cep;
FILE	*fp;

{
	register char	*cp;
	int	c;
	char	cacheline2[MAXLEN];



	if ( fgets( cep->cacheline, MAXLEN, fp) == NULL) {
		return NULL;
	}

	chop( cep->cacheline);
	*cep->content = *cep->suffix = '\0';
	cep->type1 = *(cep->cacheline);

	switch ( cep->type1) {
	case 'i':
		cep->entrytype = ILINE;
		return (cep);
	case 'h':
		if ( strncmp( cep->cacheline, "http:", 5) == 0 ) {
			cep->entrytype = HTTPTEXT;
			cep->type1 = '\0';
			return (cep);
		}
	}

	if ( ( c = fgetc( fp)) == '\t') {
		if ( fgets( cacheline2, MAXLEN, fp) == NULL) {
			senderr2( "Corrupt secondary cache line", "");
			exit( 2);
		}
		chop( cacheline2);
		if ( streq( cacheline2, "gopher_link\t"))
			cep->entrytype = REMGOPHER;
		else if ( streq( cacheline2, "gn_link\t"))
			cep->entrytype = REMGN;
		else {
			cep->entrytype = LOCAL;
			strcpy( cep->content, cacheline2);
			if ( ( cp = strchr( cep->content, '\t')) != NULL) {
				*cp++ = '\0';
				strcpy( cep->suffix, cp);
			}
		}
	}
	else {
		ungetc( c, fp);
		/* No secondary cacheline:  assume a remote gopher */
		cep->entrytype = REMGOPHER;
	}
	parse_cache( cep);
	return (cep);
}

/*
 * parse_cache( cep) looks at a line from a .cache file and fills in
 * the fields of the Cache_entry struct pointed to by cep.  These fields
 * are name,  path, host and port.  The path has ending '/' removed 
 * unless it is the root which  could have path / or 1/ or 7c/.  
 * Type1 is the single character type which is the first character of
 * the .cache line.
 */


static void
parse_cache(  cep)

Cache_entry	*cep;
{
	register char	*cp,
			*cp2,
			*cp3,
			*cp4;

	char		cline[MAXLEN];


	assert ( (*(cep->cacheline) != '\t') );
	strcpy( cline, cep->cacheline);
	
	if ( (cp = strchr( cline, '\t')) == (char *) NULL) {
		senderr2( "Corrupted cache entry (no tabs): ", cline);
		exit( 2);
	}

	*cp++ = '\0';
	strcpy( cep->name, &cline[1]);
	
	/* cp points after 1st tab */

	if ( (cp2 = strchr( cp, '\t')) == (char *) NULL) {
		senderr2( "Corrupted cache entry (no 2nd tab): ", cline);
		exit( 2);
	}
	*cp2++ = '\0';

	/* cp2 points after 2nd tab */

	if ( (*cp == '\0') || streq( cp, "1"))
		strcpy(cep->path, "1/");
	else
		strcpy( cep->path, cp);

	/* Make sure path does not end in '/' except for root */

	if (cep->path[strlen(cep->path)-1] == '/')
		cep->path[strlen(cep->path)-1] = '\0';

	if ( (cp3 = strrchr( cep->path, '/')) == NULL)	/* It's the root */
		strcat( cep->path, "/");  /* Put back '/' */

	if ( (cp3 = strchr( cp2, '\t')) == (char *) NULL) {
		senderr2( "Corrupted cache entry (no 3rd tab): ", cline);
		exit( 2);
	}
	*cp3++ = '\0';

	/* cp3 points after 3rd tab */

	strcpy( cep->host, cp2);
	if ( (cp4 = strchr( cp3, '\t')) !=  NULL)
		*cp4 = '\0';
	strcpy( cep->port,  cp3);

}


void
www_cache_line( cep)
Cache_entry	*cep;
{
	register char	*cp,
			*cp2;
	char		*prog,
			buf[PATHLEN];

	Item    	item;

	char	out[PATHLEN];


	switch ( cep->entrytype) {
	case HTTPTEXT:
		strcpy( out, cep->cacheline + 5);
		send_text_line( out);
		return;
	case ILINE:
		strcpy(out, cep->cacheline + 1);
		cp = strchr( out, '\t');
		*cp = '\0';
		send_text_line( out);
		return;
	}

	if ( accesstype == DIRCHK) {
		if ( (*(cep->path) == '1') && (cep->entrytype == LOCAL) ) {
			strcpy( buf, cep->path);
			parse_selector( &item, buf, FULL);
			if (!chkaccess( &item, NOLOG))  
				return;
		}
	}
	www_escape( cep->path);
	www_replace( cep->name);

	switch (cep->type1) {
	case '8':
	case 'T':
		strcpy( buf, cep->cacheline);
		prog = ( cep->type1 == '8' ? "telnet" : "tn3270");
		sprintf( out, "<li> <a href=\"%s://%s:%s\">%s</a>", 
			prog, cep->host, cep->port, cep->name);
		break;
	default:

		if ( (strncmp( cep->path, "ftp", 3) == 0 ) &&
				((cp2 =  strchr(cep->path, ':')) != NULL) &&
				((cp =  strchr(cp2, '@')) != NULL) ) {
			*cp++ = '\0';
			cp2++;
			sprintf( out, "<li> <a href=\"ftp://%s%s\">%s</a>", 
				cp2, cp, cep->name);
			break;
		}
		if ( cep->entrytype == LOCAL || cep->entrytype == REMGN ) {
			sprintf(out,"<li> <a href=\"http://%s:%s/%s\">%s</a>", 
				cep->host, cep->port, cep->path, cep->name);
			break;
		}
		sprintf(out, "<li> <a href=\"gopher://%s:%s/%c%s\">%s</a>", 
			cep->host, cep->port, cep->type1, 
			cep->path, cep->name);

	}
	send_text_line( out);
}




/*
 * www_escape( path) replaces special chars in path with %XX escapes
 */

static void
www_escape( path)
char	*path;
{

	register char	*cp,
			*cp2;
	char	buf[BUFSIZE];

	cp = path;
	cp2 = buf;
	while ( *cp ) {
		switch (*cp) {
		case ',':
		case ';':
		case '"':
		case '\'':
		case '&':
		case '=':
		case '(':
		case ')':
		case '{':
		case '}':
		case ' ':
			sprintf( cp2, "%%%X", (int) *cp);
			cp2 += 3;
			cp++;
			break;
		default:
			*cp2++ = *cp++;
		}
	}
	*cp2 = '\0';
	strcpy( path, buf);
}

/*
 * www_unescape( path) undoes what www_escape does.  Also change
 * any +'s to pluschar.
 */

void
www_unescape( path, pluschar)
char	*path,
	pluschar;
{

	register char	*cp,
			*cp2;
	int	val;

	char	minbuf[3],
		buf[PATHLEN];

	cp = path;
	cp2 = buf;
	while ( *cp ) {
		switch ( *cp) {
		case '%':
			cp++;
			minbuf[0] = *cp++;
			minbuf[1] = *cp++;
			minbuf[2] = '\0';
			sscanf( minbuf, "%x", &val);
			*cp2++ = (char) val;
			break;
		case '+':
			*cp2++ = pluschar;
			cp++;
			break;
		default:
			*cp2++ = *cp++;
		}
	}
	*cp2 = '\0';
	strcpy( path, buf);
}


/*
 * www_replace ( name) replaces special characters, <, >, & in a name
 */
  
static void
www_replace ( name)
char	*name;
{
 	register char	*cp,
 			*cp2;
	char	buf[BUFSIZE];

	cp = name;
	cp2 = buf;
	while ( *cp ) {
		switch (*cp) {
		case '&':
			strcpy( cp2, "&amp;");
			cp2 += 5;
			break;
		case '<':
			strcpy( cp2, "&lt;");
			cp2 += 4;
			break;
		case '>':
			strcpy( cp2, "&gt;");
			cp2 += 4;
			break;
		default:
			*cp2++ = *cp;
		}
		cp++;
	}
	*cp2 = '\0';
	strcpy( name, buf);
}
