/*
    Gn: A Server for the Internet Gopher Protocol(*).
    File: gn/util.c
    Version 1.0

    Copyright (C) 1993  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    (*) Gopher is a registered trademark of the Univ. of Minn.
*/
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <netdb.h>
#include "gn.h"


void
senderr( msg, ip)
char	*msg;
Item	*ip;
{
	senderr2( msg, ip->selector);
}

void
www_err( status, msg)
char	*status,
	*msg;
{
	char	buf[BUFSIZE];

	if ( http & HTTP1_0) {
		sprintf( buf, "%s %s %s", HTTPVERSION, status, msg);
		send_text_line( buf);
		sprintf( buf, "Server: %s", VERSION);
		send_text_line( buf);
		send_text_line( "");
	}
	fprintf( stdout, "<head><title>%s %s</title></head>\n", 
				status, msg);
	fprintf( stdout, "<body><h2>Error code %s</h2>\n%s\n</body>\n", 
				status, msg);
}

void
senderr2( msg, msg2)
char	*msg,
	*msg2;
{
	char	*status;


	if ( msg2 == NULL)
		msg2 = "";

	if ( http){
		if ( strncmp( msg, "Sorry", 5) == 0 ) {
			status = "403";
		}
		else {
			status = "400";
		}
		www_err( status, msg);
	}
	else
		printf(	"3Server error: %s\t\terror.host\t0\r\n.\r\n", msg);

	writelog( msg, msg2);
}


void
get_remote_info( ipnum, hostp)
char	*ipnum,
	*hostp;
{
	char			*inet_ntoa();
	struct sockaddr_in	saddr;
	struct hostent		*hostentp = NULL;
	int			size;

#ifdef NONET
	*ipnum = *hostp = '\0';
	return;
#endif
	size = sizeof(saddr);
	getpeername(fileno(stdin), &saddr, &size);
	strcpy(ipnum, inet_ntoa(saddr.sin_addr));

	hostentp = gethostbyaddr((char *) &saddr.sin_addr,
			sizeof (saddr.sin_addr.s_addr), AF_INET);

		if (hostentp == NULL)
			strcpy( hostp, ipnum);
		else {
			strcpy( hostp, hostentp->h_name);
			strlower( hostp);
		}
}


/* chop( line)  Cut out CRLF at end of line, or just LF */

void
chop( line)
char *line;
{
	register char	*cp;

	if ( *line == '\0')
		return;
	cp = line;
	while ( *cp )
		cp++;
	if ( *--cp == '\n') {
		*cp = '\0';
		if ( (cp > line) && *--cp == '\r')
			*cp = '\0';
	}
}

/*
 *  wild_match
 *
 *  String equality routine, including matching the '*' character.
 */

/*
 *  Borrowed from the ANU News sources V6.1b3 newsrtl.c.  Original sources
 *  Copyright 1989 by Geoff Huston.
 *
 */

int wild_match(l,p)
    char *l,
         *p;
{
    if (!*l) {
        if (!*p) return(1);
        else if (*p == '*') return(wild_match(l,p+1));
        else return(0);
        }
    if (*p == '*') {
        while (!wild_match(l,p+1)) {
            l++;
            if (!*l) {
                if (!*(p+1)) return(1);
                else return(0);
                }
            }
        return(1);
        }
    if ((*p == '%') || (*p == '?')) return(wild_match(l+1,p+1));
    if (*p == '^') return((*l == *(p+1)) && wild_match(l+1,p+2));
    return((*l == *p) && wild_match(l+1,p+1));
}

/*
 *  Make string all lower case.
 */


char *
strlower( st)
char	*st;
{
	register char	*cp;

	cp = st;
	while ( *cp) {
		*cp =  (isupper(*cp) ? *cp - 'A' + 'a' : *cp );
		cp++;
	}
	return (st);
}

FILE *
safer_popen( command, args, type)
char	*command,
	*args,
	*type;
{
	register char	*cp,
			*cp2;
	char	buf[PATHLEN];

	strcpy( buf, command);

	if ( (cp = strrchr( buf, '/')) != NULL)  {
		*cp = '\0';
		if ( chdir( buf) != 0  )
			writelog( "Can't chdir for exec to: ", buf);
		*cp = '/';
	}

	cp = args;
	if ( *cp) {
		cp2 = buf + strlen( buf);
		*cp2++ = ' ';
		while ( *cp != '\0') {
			switch (*cp) {
				case	';':
				case	'`':
				case	'\'':
				case	'|':
				case	'*':
				case	'?':
				case	'-':
				case	'~':
				case	'>':
				case	'<':
				case	'^':
				case	'[':
				case	']':
				case	'&':
				case	'(':
				case	')':
				case	'{':
				case	'}':
				case	'$':
				case	'=':
				case	'!':
				case	'/':
				case	'\\':
#ifdef ESCAPE_META
					*cp2++ = '\\';
					*cp2++ = *cp++;
#else
					cp++;
					*cp2++ = ' ';
#endif
					break;
				default:
					*cp2++ = *cp++;
			}
		}
		*cp2 = '\0';
	}
	else
		strcpy( buf, command);

	return (popen( buf, type));
}
