# g-expr.pl version 1.0
#
#  Handles parsing of kill files and searches
#
#  UNDOCUMENTED

######################################################################
# PARSE STRUCTURE:  array of ("#" denotes a tab, "..." is repetition)
#
# error     # text # ...
# gtime     # yyyymmddhhmm
# seconds   # seconds
# item      # var # op # value
# item0     # var # op # value
# depth     # value
# limit     # value
# veronica  # term # ...
#
#    term= xword | and xword | not xword
#    xword= word | word* | *word | *word*
######################################################################

#---------------------------------------------------------------------
# INITIALIZATION

@seconds_top = ( 1,               # seconds
                 60,              # minutes
                 60*60,           # hours
                 60*60*24,        # days
                 60*60*24*7,      # weeks
                 60*60*24*30,     # month
                 60*60*24*30*12   # year
               );

require "g-time.pl";

######################################################################
# KILL FILE
######################################################################

# &load_kill_file($filename) --> $code
#

sub load_kill_file {
  local($filename)= @_;
  local(@keep, @kill, @keep_subs_of, @kill_subs_of);
  local($kw, $expr, @expr, $expr_code);

  if ($filename ne "") {
    open(KILL,$filename) || die "$filename: $!\n";
    while (<KILL>) {
      chop;
      s/#.*//;                             ## remove comments,
      next if /^\s*$/;                     ## skip empty lines
      ($kw, $expr)= /^([\w-]+)\s*(.*)$/;
      $kw =~ tr/A-Z/a-z/;
      if ($kw !~ /^(keep|kill)(-subs-of)?$/) {
        die "$filename line $.: keyword '$kw' not recognized\n...Line should".
            "start with keep, kill, keep-subs-of, or kill-subs-of\n";
      }
      @expr = &parse_expr($expr);
      $error= &expr_find1("error",@expr);
      die "$filename line $.: can't parse $error\n" if $error;
      $expr_code = join(" && ", &expr_code(@expr));
      next if $expr_code eq "";
      if    ($kw eq "keep"        ) { push(@keep,         $expr_code); }
      elsif ($kw eq "kill"        ) { push(@kill,         $expr_code); }
      elsif ($kw eq "keep-subs-of") { push(@keep_subs_of, $expr_code); }
      elsif ($kw eq "kill-subs-of") { push(@kill_subs_of, $expr_code); }
      else { die "Programmer error"; }
    }
    close(KILL);
  }
  local($subs)= &make_code("KEEP", "keep", "kill");
  $subs .= &make_code("KEEP_SUBS_OF", "keep_subs_of", "kill_subs_of");
  $subs . "1;\n\n";
}

sub make_code {
  local($cname, *keep_arg, *kill_arg)= @_;
  local($sub)="\nsub $cname {\n";
  if (@keep_arg) {
    if (@kill_arg) {
      $sub .= ("    ( (". join(")\n    ||(", @keep_arg) . "))\n"
              ."&& !( (". join(")\n    ||(", @kill_arg) . "));\n}\n\n");
    }
    else {
      $sub .= ("   (". join(")\n ||(", @keep_arg). ");\n}\n\n");
    }
  }
  else {
    if (@kill_arg) {
      $sub .= ("  !((". join(")\n ||(", @kill_arg). "));\n}\n\n");
    }
    else {
      $sub .= "  1;\n}\n\n";
    }
  }
  $sub;  # return
}

########################################################################
# EXPRESSION
########################################################################

#---------------------------------------------------------------------
# Expression Parse  -- &parse_expr(text) -> parse_structure

sub parse_expr {
  local($text)= @_;
  local($parse_gtime)= "000000000000";
  local($parse_seconds, $parse_error, $error);
  $text =~ tr/\t/ /;

  @expr = ();

  $text =~ s/^\s+//;
  while($text ne "") {
    if    (&is_item(     $text)) { ($text, $error)= &parse_item(     $text); }
    elsif (&is_parameter($text)) { ($text, $error)= &parse_parameter($text); }
    elsif (&is_veronica( $text)) { ($text, $error)= &parse_veronica( $text); }
    elsif (&is_year(     $text)) { ($text, $error)= &parse_year(     $text); }
    elsif (&is_date1(    $text)) { ($text, $error)= &parse_date1(    $text); }
    elsif (&is_date2(    $text)) { ($text, $error)= &parse_date2(    $text); }
    elsif (&is_hour(     $text)) { ($text, $error)= &parse_hour(     $text); }
    elsif (&is_seconds(  $text)) { ($text, $error)= &parse_seconds(  $text); }
    else                         { ($text, $error)= ("", $text);            }
    if ($error) {
      $error =~ /^([^&]*)(.*)/;  $parse_error .= "\t$1";  $text= $2;
    }
    $text =~ s/^\s*&+\s*//;
  }
  unshift(@expr,"seconds\t$parse_seconds") if $parse_seconds > 0;
  unshift(@expr,"gtime\t$parse_gtime")     if $parse_gtime ne "000000000000";
  unshift(@expr,"error$parse_error")       if $parse_error ne "";
  return @expr
}

#---------------------------------------------------------------------
# Expression Code

sub expr_code {
  local($_, $class,@code, $c);
  for (@_) {
    @F= split(/\t/, $_);
    $class= shift(@F);
    if    ($class eq "error"    ) {}
    elsif ($class eq "gtime"    ) { $c= &gtime_code(   @F); }
    elsif ($class eq "seconds"  ) { $c= &seconds_code( @F); }
    elsif ($class eq "item"     ) { $c= &item_code(    @F); }
    elsif ($class eq "item0"    ) { $c= &item_code(    @F); }
    elsif ($class eq "depth"    ) { $c= &depth_code(   @F); }
    elsif ($class eq "limit"    ) { $c= &limit_code(   @F); }
    elsif ($class eq "veronica" ) { $c= &veronica_code(@F); }
    else { die "Programmer error, no '$class' class!"; }
    push(@code, $c) if $c;
  }
  @code;
}

#---------------------------------------------------------------------
# Expression Nice

sub expr_nice {
  local($_, $class, @nice, $c);
  for (@_) {
    @F= split(/\t/, $_);
    $class= shift(@F);
    if    ($class eq "error"    ) {}
    elsif ($class eq "gtime"    ) { $c= &gtime_nice(   @F); }
    elsif ($class eq "seconds"  ) { $c= &seconds_nice( @F); }
    elsif ($class eq "item"     ) { $c= &item_nice(    @F); }
    elsif ($class eq "item0"    ) { $c= &item_nice(    @F); }
    elsif ($class eq "depth"    ) { $c= &depth_nice(   @F); }
    elsif ($class eq "limit"    ) { $c= &limit_nice(   @F); }
    elsif ($class eq "veronica" ) { $c= &veronica_nice(@F); }
    else { die "Programmer error, no '$class' class!"; }
    push(@nice, $c) if $c;
  }
  join(" && ", @nice);
}

########################################################################
# Search parse_structure

sub expr_find1 {
  local($class,@expr)= @_;
  local($_,$c,$r);
  for (@expr) {
    ($c,$r)= split(/\t/,$_,2);
    return $r if $c eq $class;
  }
  return undef;
}

sub expr_find1rem {
  local($class,@expr)= @_;
  local($_,$c,$r);
  for (0..$#expr) {
    ($c,$r)= split(/\t/,$expr[$_],2);
    return ($r, @expr[0..$_-1], @expr[$_+1..$#expr]) if $c eq $class;
  }
  return (undef, @expr);
}

########################################################################
# ITEM (var op value)
########################################################################

#---------------------------------------------------------------------
# Parse LVALUE OP VALUE
#    LVALUE ::=  (type | name | path | host | port ) 0?
#    OP     ::=  = | != | ~ | =~ | !~ | eq | ne
#    VALUE  ::=  word | "text" | 'text'
# -- push(@expr,"item0?\tLVALUE\tNORMAL_OP\tVALUE")
#    NORMAL_OP ::=  eq | ne | =~ | !~

sub is_item {
  local($_)= @_;
  /^(type|name|path|host|port)0?\b/i;
}

sub parse_item {
  local($text) = @_;
  local($kw, $op, $value);
  ($kw,   $rest) = $text =~ /^(\w+)\s*(.*)/;
  ($op,   $rest) = &get_op($rest);    return ("", $rest) unless defined $op;
  ($value,$rest) = &get_value($rest); return ("", $rest) unless defined $value;
  $kw =~ tr/A-Z/a-z/;
  push (@expr, "item\t$kw\t$op\t$value");
  return ($rest, "");
}

sub get_op {
  local($_)= @_;
  local($op,$rest) = /^(=~?|![=~]|~|eq|ne)\s*(.*)/i;
  return (undef, $_) if $op eq "";
  if    ($op eq "=" ) { $op="eq"; }
  elsif ($op eq "!=") { $op="ne"; }
  elsif ($op eq "~" ) { $op="=~"; }
  return ($op, $rest);
}

sub get_value {
  local($_)= @_;
  return ($1, $')  if  /^"([^"]+)"\s*/;  # "..." quoted string
  return ($1, $')  if  /^'([^']+)'\s*/;  # '...' quoted string
  return ($1, $')  if  /^([^"']\S*)\s*/; # single word
  return (undef, $_);
}

#---------------------------------------------------------------------
# Item Code

sub item_code {
  local($lval, $op, $val)= @_;
  if (substr($op,-1) eq "~") {
    $val =~ s:/:\\/:g;
    "(\$$lval $op /$val/)";
  }
  else {
    $val =~ s:':\\':g;
    "(\$$lval $op '$val')";
  }
}

#---------------------------------------------------------------------
# Item Nice

sub item_nice {
  local($var,$op,$value) = @_;
  "$var $op $value";
}

########################################################################
# PARAMETER (var = value)
########################################################################

#---------------------------------------------------------------------
# Parameter Parse
#
#   limit OP? natural    -- push(@expr,"limit\tnatural")
#   depth OP? natural    -- push(@expr,"depth\tnatural")
#    OP ::=  = | < | <=

sub is_parameter {   # limit=\d+ || depth=\d+
  local($_)= @_;
  /^limit\b/i || /^depth\b/i;
}

sub parse_parameter {
  local($_)= @_;
  local($kw,$val,$rest)= /^(\w+)\s*[=<]*\s*(\d+)\s*(.*)/;
  return ("", $_) if $val eq "";
  $kw =~ tr/A-Z/a-z/;
  push(@expr, "$kw\t$val");
  return ($rest,"");
}

#---------------------------------------------------------------------
# Parameter Code

sub limit_code {
  local($limit)= @_;
  "(\$limit <= $limit)";
}

sub depth_code {
  local($depth)= @_;
  "(\$depth <= $depth)";
}

#---------------------------------------------------------------------
# Parameter Nice

sub limit_nice {
  local($value)= @_;
  "limit=$value";
}

sub depth_nice {
  local($value)= @_;
  "depth=$value";
}

########################################################################
# VERONICA
########################################################################

#---------------------------------------------------------------------
# Veronica Parse
#   VERONICA  ::= veronica VERA_TERM*
#   VERA_TERM ::=  WORD | and WORN | not WORD | or WORD
#   WORD ::= '*'? word '*'?
# -- push(@expr, "veronica\t@vera_terms") (terms normalised, separated by "\t")

sub is_veronica {  ## veronica ...
  local($_)= @_;
  /^veronica\b/i;
}

sub parse_veronica {
  local($text) = @_;
  local($rest);

  $text =~ s/^veronica\s*//i;
  ($text,$rest)= ($1,$2)  if  $text =~ /^([^&]*)&+\s*(.*)/;

  @terms = $text =~ /(and\s*\S+|or\s*\S+|\S+)/g;
  for (@terms) {
    s/^or\s+//;
    return ("",$text) if $_ eq "and" || $_ eq "or" || $_ eq "not";
  }
  push(@expr, "veronica\t". join("\t", @terms));
  return ($rest, "");
}

#---------------------------------------------------------------------
# Veronica Code

sub veronica_code {
  local($vera_code, $_, $conj, $op);
  for (@_) {
    $conj="||"; $op="=~";
    if    (s/^and\s*//) { $conj= "&&"; }
    elsif (s/^not\s*//) { $conj= "&&"; $op= "!~"; }
    s:\W:\\$1:;
    if (substr($_,0,2) eq '\*') { $_ = substr($_,2); }
    else                        { $_ = ('\b'.$_); }
    if (substr($_,-2)  eq '\*') { $_ = substr($_,0,length($_)-2); }
    else                        { $_ .= '\b'; }
    $vera_code= "($vera_code) $conj "  if $vera_code;
    $vera_code .= "\$name $op /$_/";
  }
  $vera_code;
}

#---------------------------------------------------------------------
# Veronica Nice

sub veronica_nice {
  local(@Vterms)= @_;
  if (@Vterms) { "veronica ".join(" ",@Vterms); }
  else         { ""; }
}

########################################################################
# TIME
########################################################################

#---------------------------------------------------------------------
# Parse  year  (4 digits, 19xx or 20xx)  -- change $parse_gtime

sub is_year {    ## \d\d\d\d
  local($_)= @_;
  /^19\d\d\b/ || /^20\d\d\b/;
}

sub parse_year {
  local($_)= @_;
  s/^(\d\d\d\d)\s*//;
  substr($parse_gtime,0,4)= $1;
  return($',"");
}

#---------------------------------------------------------------------
# Parse date:   -- change $parse_gtime
#  1)  DAY MONTH
#  2)  MONTH DAY
#  DAY ::=  natural                       (1<= day <= 31)
#  MONTH ::= (jan | ... | dec) letter*    (only the three first letters matter)

sub is_date1 {   ## \d+ Month
  local($_)= @_;
  /^(\d\d?)\s*(\w\w\w)\w*/;
  $1>0 && $1<=31
       && ($2 =~ /^(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)/i);
}

sub parse_date1 {
  local($_)= @_;
  ($d,$m,$rest) = /^(\d+)\s*(\w\w\w)\w*\s*(.*)/;
  $m =~ tr/A-Z/a-z/;
  $m = index(".jan.feb.mar.apr.may.jun.jul.aug.sep.oct.nov.dec", ".$m") /4 +1;
  substr($parse_gtime,4,4)= sprintf("%.2d%.2d",$m,$d);
  return ($rest,"");
}

sub is_date2 {   ## Month \d+
  local($_)= @_;
  /^(\w\w\w\w*)\s*(\d\d?)/;
  $2>0 && $2<=31
       && ($1 =~ /^(jan|feb|mar|apr|may|jun|jul|aug|sep|oct|nov|dec)/i);
}

sub parse_date2 {
  local($_)= @_;
  ($m,$d,$rest) = /^(\w\w\w)\w*\s*(\d+)\s*(.*)/;
  $m =~ tr/A-Z/a-z/;
  $m = index(".jan.feb.mar.apr.may.jun.jul.aug.sep.oct.nov.dec", ".$m") /4 +1;
  substr($parse_gtime,4,4)= sprintf("%.2d%.2d",$m,$d);
  return ($rest,"");
}

#---------------------------------------------------------------------
# Parse HH:MM  -- change $parse_gtime

sub is_hour {   ## \d\d:\d\d
  local($_) = @_;
  /^\d\d?:\d\d?\b/;
}

sub parse_hour {
  local($_)= @_;
  local($h,$m,$rest)= /^(\d\d?):(\d\d?)\s*/;
  substr($parse_gtime,8,4)= sprintf("%.2d%.2d",$h,$m);
  return($rest, "");
}

#---------------------------------------------------------------------
# Time Code

sub gtime_code {
  local($gtime)= @_;
  $gtime= &absolute_gtime($gtime);
  "(\$gtime gt '$gtime')";
}

#---------------------------------------------------------------------
# Time Nice

sub gtime_nice {
  local($gtime)= @_;
  &gtime2text($gtime);
}

########################################################################
# SECONDS
########################################################################

#---------------------------------------------------------------------
# Parse  natural TIME_UNIT  -- change $parse_seconds
#  TIME_UNIT ::= years | months | days | hours | minutes | seconds

sub is_seconds {  ## \d+ unit
  local($_)= @_;
  /^(\d+)\s*(\w+)/;
  &time_unit($2."");
}

sub parse_seconds {
  local($_)= @_;
  local($n,$u,$rest)= /^(\d+)\s*(\w+)\s*(.*)/;
  $parse_seconds += ($n*&time_unit($u));
  return ($rest, "");
}

sub time_unit {
  local($_)= @_;
  local($n)=length($_); tr/A-Z/a-z/;
  if ($_ eq substr("seconds",0,$n)) { return $seconds_top[0]; }
  if ($_ eq substr("minutes",0,$n)) { return $seconds_top[1]; }
  if ($_ eq substr("hours",  0,$n)) { return $seconds_top[2]; }
  if ($_ eq substr("days",   0,$n)) { return $seconds_top[3]; }
  if ($_ eq substr("weeks",  0,$n)) { return $seconds_top[4]; }
  if ($_ eq substr("months", 0,$n)) { return $seconds_top[5]; }
  if ($_ eq substr("years",  0,$n)) { return $seconds_top[6]; }
  return 0;
}

#---------------------------------------------------------------------
# Seconds Code

sub seconds_code {
  local($seconds)= @_;
  local($gtime)= &time2gtime(time()-$seconds);
  "(\$gtime gt '$gtime')";
}

#---------------------------------------------------------------------
# Seconds Nice

sub seconds_nice {
  local($seconds) = @_;
  local($nice)="";
  for $i ((6, 5, 4, 3, 2, 1, 0)) {
    if ($seconds>=$seconds_top[$i]) {
      $v=int($seconds/$seconds_top[$i]);
      $seconds -= ($v * $seconds_top[$i]);
      $nice .= " "  if  $nice ne "";
      $nice .= ($v . " ");
      $nice .= ("second", "minute", "hour", "day", "week", "month", "year")[$i];
      $nice .= "s"  if  $v>1;
    }
  }
  $nice;
}

########################################################################
# End

1; # for require
